using System;
using System.Drawing;

namespace PDim2Rigid {
    public struct Dim2Vector {

        public float x, y;

        public Dim2Vector(float ax, float ay) {
            x = ax;
            y = ay;
        }

        public static Dim2Vector operator+(Dim2Vector a, Dim2Vector b) {
            Dim2Vector tmp;
            tmp.x = a.x + b.x;
            tmp.y = a.y + b.y;
            return tmp;
        }

        public static Dim2Vector operator-(Dim2Vector a, Dim2Vector b) {
            Dim2Vector tmp;
            tmp.x = a.x - b.x;
            tmp.y = a.y - b.y;
            return tmp;
        }

        public static Dim2Vector operator-(Dim2Vector a) {
            Dim2Vector tmp;
            tmp.x = -a.x;
            tmp.y = -a.y;
            return tmp;
        }

        public static float operator*(Dim2Vector a, Dim2Vector b) {
            return a.x * b.x + a.y * b.y;
        }

        public static Dim2Vector operator*(Dim2Vector a, float f) {
            Dim2Vector tmp;
            tmp.x = a.x * f;
            tmp.y = a.y * f;
            return tmp;
        }

        public static Dim2Vector operator*(float f, Dim2Vector a) {
            Dim2Vector tmp;
            tmp.x = a.x * f;
            tmp.y = a.y * f;
            return tmp;
        }

        public static Dim2Vector operator/(Dim2Vector a, float f) {
            Dim2Vector tmp;
            tmp.x = a.x / f;
            tmp.y = a.y / f;
            return tmp;
        }

        public static float operator%(Dim2Vector a, Dim2Vector b) {
            return a.x * b.y - a.y * b.x;
        }

        public static Dim2Vector operator%(Dim2Vector a, float f) {
            Dim2Vector tmp;
            tmp.x = a.y * f;
            tmp.y = -a.x * f;
            return tmp;
        }

        public static Dim2Vector operator%(float f, Dim2Vector a) {
            Dim2Vector tmp;
            tmp.x = -a.y * f;
            tmp.y = a.x * f;
            return tmp;
        }

        public Point ToPoint() {
            return new Point((int)x, (int)y);
        }

        public float norm() {
            return (float)Math.Sqrt(x * x + y * y);
        }

        public float sqr() {
            return x * x + y * y;
        }

        public Dim2Vector cross() {
            Dim2Vector tmp;
            tmp.x = y;
            tmp.y = -x;
            return tmp;
        }

        public Dim2Vector nmlz() {

            Dim2Vector tmp;

            float r = (float)Math.Sqrt(x * x + y * y);
            if (r == 0) {
                tmp.x = x;
                tmp.y = y;
            }
            else {
                tmp.x = x / r;
                tmp.y = y / r;
            }
            return tmp;
        }



        //̑
        public static bool lineHitTest(Dim2Vector l11, Dim2Vector l12,
                Dim2Vector l21, Dim2Vector l22, out Dim2Vector p) {

            p = new Dim2Vector();

            Dim2Vector v1 = l12 - l11;
            Dim2Vector v2 = l22 - l21;
            float c = v2 % v1;

            if (c == 0) {
                return false;
            }

            Dim2Vector v3 = l21 - l11;
            float a = (v2 % v3) / c;

            if ((0 <= a) && (a <= 1)) {

                double b = (v1 % v3) / c;

                if ((0 <= b) && (b <= 1)) {
                    p = l11 + v1 * a;
                    return true;
                }
            }

            return false;
        }

        public static bool isConvex(Dim2Vector[] verticies) {
            if (verticies.Length < 3) {
                return false;
            }
            int j, k;
            for (int i = verticies.Length-1; i >= 0; i--) {
                j = i + 1;
                k = i + 2;
                if (j >= verticies.Length) j -= verticies.Length;
                if (k >= verticies.Length) k -= verticies.Length;

                if ((verticies[i] % verticies[j]) < 0) {
                    return false;
                }
            }

            return true;
        }

    }

}