﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.Drawing;


namespace PDim2Rigid {
    public class GDIFunctions {
        public const int DC_BRUSH = 18;
        public const int DC_PEN = 19;

        [DllImport("user32.dll")]
        public static extern int FillRect(int hDC, ref Rectangle lprc,
                                           int hbr);
        [DllImport("user32.dll")]
        public static extern int GetDC(int hWnd);

        [DllImport("gdi32.dll")]
        public static extern int CreateSolidBrush(int color);

        [DllImport("gdi32.dll")]
        public static extern int CreatePen(int style, int width, int color);

        [DllImport("gdi32.dll")]
        public static extern int DeleteObject(int obj);
        
        [DllImport("gdi32.dll")]
        public static extern int Polygon(int hdc, Point[] pt, int nCount);

        [DllImport("gdi32.dll")]
        public static extern int SelectObject(int hdc, int obj);

        [DllImport("gdi32.dll")]
        public static extern int SetDCBrushColor(int hdc, int color);

        [DllImport("gdi32.dll")]
        public static extern int SetDCPenColor(int hdc, int color);

        [DllImport("gdi32.dll")]
        public static extern int GetStockObject(int type);

        [DllImport("gdi32.dll")]
        public static extern int MoveToEx(int hdc, int x, int y, ref Point pt);

        [DllImport("gdi32.dll")]
        public static extern int LineTo(int hdc, int x, int y);

        [DllImport("gdi32.dll")]
        public static extern int CreateCompatibleDC(int hdc);

        [DllImport("gdi32.dll")]
        public static extern int CreateCompatibleBitmap(int hdc, int width, int height);

        [DllImport("gdi32.dll")]
        public static extern int DeleteDC(int hdc);

        [DllImport("user32.dll")]
        public static extern int ReleaseDC(int hWnd, int hdc);

        [DllImport("gdi32.dll")]
        public static extern int BitBlt(int hdcDest, int xDest, int yDest, int width, int height,
                int hdcSrc, int xSrc, int ySrc, int op);



        public static int ColorToInt(Color color) {
            int a = color.ToArgb();
            int r, g, b;
            r = (a >> 16) & 255;
            g = (a >> 8) & 255;
            b = a & 255;

            return r + ((g + (b << 8)) << 8);
        }

    }

    public class GDIBackBuffer : IDisposable {
        private int dc;
        private int bmp;
        public GDIBackBuffer(int hdc, int width, int height) {
            dc = GDIFunctions.CreateCompatibleDC(hdc);
            bmp = GDIFunctions.CreateCompatibleBitmap(hdc, width, height);
            GDIFunctions.SelectObject(dc, bmp);
        }
        void IDisposable.Dispose() {
            GDIFunctions.DeleteDC(dc);
            GDIFunctions.DeleteObject(bmp);
        }
        public int DCHandle {
            get {
                return dc;
            }            
        }
        public int BitmapHandle {
            get {
                return bmp;
            }            
        }
    }

    public class GDIBrush : IDisposable {
        private int br;
        public GDIBrush(Color color) {
            br = GDIFunctions.CreateSolidBrush(GDIFunctions.ColorToInt(color));
        }
        void IDisposable.Dispose() {
            GDIFunctions.DeleteObject(br);
        }

        public int Handle {
            get {
                return br;
            }
            set {
                br = value;
            }
        }
    }

    public class GDIPen : IDisposable {
        private int pn;
        public GDIPen(Color color, int width) {
            pn = GDIFunctions.CreatePen(0, width, GDIFunctions.ColorToInt(color));
        }
        void IDisposable.Dispose() {
            GDIFunctions.DeleteObject(pn);
        }

        public int Handle {
            get {
                return pn;
            }
            set {
                pn = value;
            }
        }
    }
        
}
