
using System;
using System.Collections.Generic;
using System.Drawing;
using System.Windows.Forms;

namespace PDim2Rigid {    

    public class RigidBodies {

        public double tm1 = 0, tm2 = 0, tm3 = 0, tm4 = 0;

        static private Color[] colorTable = new Color[10] {
            Color.Lime, Color.Blue, Color.Red, Color.Yellow, Color.Green,
            Color.Cyan, Color.Magenta, Color.Brown, Color.Gray, Color.Purple
        };

        public float ImpulseConst = 0.2f;
        public float FricConst = 1.0f;

        public float dt = 0.2f;
        public float devidingSpeed = 5.2f;
        public int maxGroupCount = 20;

        private Random random = new Random();

        private List<RigidBody> bodies = new List<RigidBody>();
        private List<HitDataShort> hits = new List<HitDataShort>();
        private ForceMatrix forceMat = new ForceMatrix();
        public List<JointData> joints = new List<JointData>();
        public List<MoterData> moters = new List<MoterData>();

        public RigidGroupList groups = new RigidGroupList(); 

        public RigidBody Pick(Dim2Vector point) {
            for (int i = bodies.Count-1; i >= 0; i--) {
                if (bodies[i].rm != 0) {
                    if (bodies[i].IsInPoint(point))
                        return bodies[i];
                }
            }
            return null;
        }

        public void Clear() {
            joints.Clear();
            bodies.Clear();
            hits.Clear();
            forceMat.clear();
        }

        public void Add(RigidBody body) {
            body.owner = this;
            body.index = bodies.Count;
            bodies.Add(body);
        }

        public void AddJoint(JointData j) {
            joints.Add(j);
        }

        public void AddJoint(RigidBody rb1, RigidBody rb2, Dim2Vector r1, Dim2Vector r2) {
            JointData j = new JointData();
            j.r1 = r1;
            j.r2 = r2;
            j.rb1 = rb1;
            j.rb2 = rb2;
            joints.Add(j);
        }

        public void addHitData(HitDataShort hds) {

            hds.hdl.hds = hds;

            int flag = 0;
            for (int i = hds.rb1.hits.Count - 1; i >= 0; i--) {
                if (hds.hdl.isEqual(hds.rb1.hits[i])) {
                    hds.hdl = hds.rb1.hits[i];
                    hds.hdl.setMaxAge(hds);
                    flag = 1;
                    break;
                }
            }
            if (flag == 0) {
                hds.rb1.hits.Add(hds.hdl);
                hds.rb2.hits.Add(hds.hdl);
            }
            hds.index = hits.Count;
            hits.Add(hds);
        }

        private bool isGroup(RigidBody rb1, RigidBody rb2) {
            if (rb1.rm == 0 || rb2.rm == 0) {
                return false;
            }
            return (((rb1.nextVel - rb2.nextVel).norm() < devidingSpeed)
                && (Math.Max(rb1.rm / rb2.rm, rb2.rm / rb1.rm) < 4.1));
        }

        private void devideToGroupsNest(List<RigidBody> ls, RigidBody body) {


            body.group = groups[groups.Count-1];
            groups[groups.Count-1].bodies.Add(body);
            ls.Remove(body);

            if (groups[groups.Count-1].bodies.Count >= maxGroupCount) {
                return;
            }
            //͓̕ւƐmɂȂ邪܂Ƃ܂肪Ȃ
            HitDataLong hdl;
            for (int i = body.hits.Count-1; i >= 0; i--) {
                hdl = body.hits[i];
                if (hdl.isMaxAge()) {
                    if (isGroup(hdl.rb1, hdl.rb2)) {
                        if (hdl.rb1 == body) {
                            for (int j = ls.Count-1; j >= 0; j--) {
                                if (hdl.rb2 == ls[j]) {
                                    devideToGroupsNest(ls, ls[j]);
                                    break;
                                }
                            }
                        }
                        else {
                            for (int j = ls.Count-1; j >= 0; j--) {
                                if (hdl.rb1 == ls[j]) {
                                    devideToGroupsNest(ls, ls[j]);
                                    break;
                                }
                            }
                        }
                    }
                }
            }
        }        

        public void devideToGroups() {
            List<RigidBody> ls = new List<RigidBody>();
            for (int i = bodies.Count-1; i >= 0; i--) {
                ls.Add(bodies[i]);
            }

            groups.Clear();

            RigidGroup gp;

            for (int i = bodies.Count-1; i >= 0; i--) {
                gp = null;
                for (int j = ls.Count-1; j >= 0; j--) {
                    if (bodies[i] == ls[j]) {
                        gp = new RigidGroup();
                        break;
                    }
                }
                if (gp != null) {
                    if (groups.Count >= colorTable.Length)
                        gp.color = Color.Black;
                    else
                        gp.color = colorTable[groups.Count % colorTable.Length];
                    groups.Add(gp);
                    devideToGroupsNest(ls, bodies[i]);
                }
            }

            groups.calculate();         
        }


        private void calcVelocities() {

            if (forceMat.col() == 0) {
                return;
            }

            DateTime dtime1, dtime2;
            dtime1 = DateTime.Now;

            int i, n;
            HitDataShort hds;

            Dim2Vector vecN;
            

            float[] a = new float[forceMat.col()];
            float[] x = new float[forceMat.col()];
            float[] c = new float[forceMat.col()];

            n = 2 * hits.Count;
            RigidBody rb1, rb2;
            for (i = 0; i < n; i++) {
                hds = hits[i / 2];
                rb1 = hds.rb1;
                rb2 = hds.rb2;

                if (i % 2 == 0) {
                    c[i] = hds.impulseConst;
                }
                else {
                    c[i] = hds.fricConst;
                }

                if (i % 2 == 0) {
                    x[i] = hds.hdl.vel;
                }
                else {
                    x[i] = hds.hdl.fric;
                } //x[i] = 0;                

                if (i % 2 == 0) {
                    float c1 = Math.Max(hds.h-1.1f, 0);
                    a[i] = rb1.velRel(rb2, hds.r1, hds.r2) * hds.n - c1 * 4.1f;
                }
                else {
                    a[i] = rb1.velRel(rb2, hds.r1, hds.r2) * hds.fn;
                }

            }

            dtime2 = DateTime.Now;
            tm1 += dtime2.Subtract(dtime1).TotalMilliseconds;
            dtime1 = dtime2;

            switch (Form1.calcMode) {
                case 0:                    
                    break;
                case 1:
                    groups.forceMat = forceMat;
                    groups.solveLarge(hits, x);
                    groups.solveSmall(a, x, c);
                    break;
                default:
                    groups.forceMat = forceMat;
                    groups.solveSmall(a, x, c);
                    break;
            } 
            
            int[] indicies = new int[forceMat.col()/2];
            for (i = indicies.Length - 1; i >= 0; i--) {
                indicies[i] = i;
            }

            MatrixMath.solveSouhosei(forceMat, indicies, a, x, c);

            dtime2 = DateTime.Now;
            tm2 += dtime2.Subtract(dtime1).TotalMilliseconds;
            dtime1 = dtime2;

            for (i = 0; i < n; i++) {
                hds = hits[i / 2];                
                
                if (i % 2 == 0) {
                    hds.hdl.vel = x[i];
                    vecN = hds.n;
                }
                else {
                    hds.hdl.fric = x[i];
                    vecN = hds.fn;
                }

                rb1 = hds.rb1;
                rb2 = hds.rb2;
                rb1.vel += vecN * (x[i] * rb1.rm);
                rb2.vel -= vecN * (x[i] * rb2.rm);
                rb1.omega += (hds.r1 % vecN) * x[i] * rb1.rI;
                rb2.omega -= (hds.r2 % vecN) * x[i] * rb2.rI;

            }

            dtime2 = DateTime.Now;
            tm1 += dtime2.Subtract(dtime1).TotalMilliseconds;
            dtime1 = dtime2;

        }

        private void calcImpulses() {

            DateTime dtime1, dtime2;
            dtime1 = DateTime.Now;

            int i, n, m, k;
            ForceMatrixItem fmi;
            HitDataLong hdl;
            HitDataShort hds;

            Dim2Vector vecN;

            forceMat.clear();
            forceMat.setSize(2 * hits.Count, 0);


            n = 2 * hits.Count;
            for (i = 0; i < n; i++) {

                hds = hits[i / 2];

                if (i % 2 == 0) {
                    vecN = hds.n;
                }
                else {
                    vecN = hds.fn;
                }

                RigidBody rb1, rb2;
                rb1 = hds.rb1;
                rb2 = hds.rb2;

                fmi = new ForceMatrixItem();
                fmi.r = i;
                fmi.value = (rb1.velByImpulse(vecN, hds.r1) + rb2.velByImpulse(vecN, hds.r2)) * vecN;
                if (fmi.value <= 0.0) {
                    MessageBox.Show("[Ŋ");
                    Console.Write("fmi.value <= 0\n");
                }
                forceMat.addRow(i, fmi);


                m = 2 * rb1.hits.Count;
                for (k = 0; k < m; k++) {

                    hdl = rb1.hits[k / 2];

                    if (hdl.hds == null)
                        continue;

                    fmi = new ForceMatrixItem();

                    if (k % 2 == 0) {
                        fmi.r = 2 * hdl.hds.index;
                    }
                    else {
                        fmi.r = 2 * hdl.hds.index + 1;
                    }

                    if (fmi.r == i)
                        continue;

                    if (rb1 == hdl.rb1) {
                        if (k % 2 == 0) {
                            fmi.value = rb1.velByImpulse(hdl.hds.n, hdl.hds.r1, hds.r1) * vecN;
                        }
                        else {
                            fmi.value = rb1.velByImpulse(hdl.hds.fn, hdl.hds.r1, hds.r1) * vecN;
                        }

                    }
                    else {
                        if (k % 2 == 0) {
                            fmi.value = -rb1.velByImpulse(hdl.hds.n, hdl.hds.r2, hds.r1) * vecN;
                        }
                        else {
                            fmi.value = -rb1.velByImpulse(hdl.hds.fn, hdl.hds.r2, hds.r1) * vecN;
                        }

                    }
                    forceMat.addRow(i, fmi);
                }

                m = 2 * rb2.hits.Count;
                for (k = 0; k < m; k++) {

                    hdl = rb2.hits[k / 2];

                    if (hdl.hds == null)
                        continue;

                    fmi = new ForceMatrixItem();

                    if (k % 2 == 0) {
                        fmi.r = 2 * hdl.hds.index;
                    }
                    else {
                        fmi.r = 2 * hdl.hds.index + 1;
                    }


                    if (fmi.r == i)
                        continue;

                    if (rb2 == hdl.rb2) {
                        if (k % 2 == 0) {
                            fmi.value = rb2.velByImpulse(hdl.hds.n, hdl.hds.r2, hds.r2) * vecN;
                        }
                        else {
                            fmi.value = rb2.velByImpulse(hdl.hds.fn, hdl.hds.r2, hds.r2) * vecN;
                        }


                    }
                    else {
                        if (k % 2 == 0) {
                            fmi.value = -rb2.velByImpulse(hdl.hds.n, hdl.hds.r1, hds.r2) * vecN;
                        }
                        else {
                            fmi.value = -rb2.velByImpulse(hdl.hds.fn, hdl.hds.r1, hds.r2) * vecN;
                        }

                    }
                    forceMat.addRow(i, fmi);
                }

            }

            if (forceMat.col() != 0) {

                float[] a = new float[forceMat.col()];
                float[] x = new float[forceMat.col()];
                float[] c = new float[forceMat.col()];

                n = 2 * hits.Count;
                RigidBody rb1, rb2;
                for (i = 0; i < n; i++) {
                    hds = hits[i / 2];

                    if (i % 2 == 0) {
                        vecN = hds.n;
                        c[i] = hds.impulseConst;
                    }
                    else {
                        vecN = hds.fn;
                        c[i] = hds.fricConst;
                    }

                    rb1 = hds.rb1;
                    rb2 = hds.rb2;

                    x[i] = 0.0F;

                    a[i] = rb1.velRel(rb2, hds.r1, hds.r2) * vecN;                    


                }

                dtime2 = DateTime.Now;
                tm3 += dtime2.Subtract(dtime1).TotalMilliseconds;
                dtime1 = dtime2;

                MatrixMath.solveImpulse(forceMat, a, x, c);

                dtime2 = DateTime.Now;
                tm4 += dtime2.Subtract(dtime1).TotalMilliseconds;
                dtime1 = dtime2;

                for (i = 0; i < n; i++) {

                    hds = hits[i / 2];

                    if (i % 2 == 0) {
                        vecN = hds.n;
                    }
                    else {
                        vecN = hds.fn;
                    }

                    rb1 = hds.rb1;
                    rb2 = hds.rb2;
                    rb1.vel += vecN * (x[i] * rb1.rm);
                    rb2.vel -= vecN * (x[i] * rb2.rm);
                    rb1.omega += (hds.r1 % vecN) * x[i] * rb1.rI;
                    rb2.omega -= (hds.r2 % vecN) * x[i] * rb2.rI;

                }


            }

            dtime2 = DateTime.Now;
            tm3 += dtime2.Subtract(dtime1).TotalMilliseconds;
            dtime1 = dtime2;
        }

        public void move1() {

            for (int i = bodies.Count - 1; i >= 0; i--) {
                (bodies[i]).move1();

                if (bodies[i].pos.x < -1000 || bodies[i].pos.x > 2000
                    || bodies[i].pos.y < -1000 || bodies[i].pos.y > 2000) {
                    bodies.RemoveAt(i);
                }
            }


            initForces();

        }

        public void setHitData() {
            hits.Clear();

            RigidBody rb1, rb2;

            List<JointData> noHitTestList = new List<JointData>();

            for (int i = joints.Count - 1; i >= 0; i--) {
                if (joints[i].doHitTest == false) {
                    noHitTestList.Add(joints[i]);
                }
            }

            for (int i = bodies.Count - 1; i >= 0; i--) {
                rb1 = (RigidBody)bodies[i];
                for (int j = i - 1; j >= 0; j--) {
                    rb2 = (RigidBody)bodies[j];
                    for (int k = noHitTestList.Count-1; k >= 0; k--) {
                        JointData nhj = noHitTestList[k];
                        if ((rb1 == nhj.rb1 && rb2 == nhj.rb2) 
                            || (rb1 == nhj.rb2 && rb2 == nhj.rb1)) {
                            rb2 = null;
                            break;
                        }
                    }
                    if (rb2 != null) {
                        rb1.hitTest(rb2);
                    }
                }
            }

            for (int i = joints.Count - 1; i >= 0; i--) {
                HitDataShort hd2 = new HitDataShort();
                hd2.hdl = new HitDataLong();
                hd2.hdl.rb1 = joints[i].rb1;
                hd2.hdl.rb2 = joints[i].rb2;
                hd2.hdl.edge1 = 1000+i;
                hd2.hdl.edge2 = 1000+i;
                hd2.r1 = joints[i].rb1.rotateVec(joints[i].r1);
                hd2.r2 = joints[i].rb2.rotateVec(joints[i].r2);

                hd2.impulseConst = -0.8f;

                Dim2Vector n;
                n = -1 * (joints[i].rb1.transformVec(joints[i].r1) 
                    - joints[i].rb2.transformVec(joints[i].r2));
                float r = n.norm();
                if (r != 0)
                    n = n.nmlz();
                else {
                    n.x = 0;
                    n.y = 1;
                }
                hd2.n = n;
                hd2.fn = n.cross();
                /*if (r >= 3.5) 
                    hd2.h = 3.5f;
                else {
                    hd2.h = r;
                }*/
                hd2.h = r * 0.1f;

                addHitData(hd2);
            }

            devideToGroups();
        }

        public void move2() {

            setHitData();            

            calcImpulses();

            RigidBody rb1, rb2;

            for (int i = moters.Count-1; i >= 0; i--) {
                rb1 = moters[i].rb1;
                rb2 = moters[i].rb2;       
                rb1.moment += moters[i].moment;
                rb2.moment -= moters[i].moment;               
            }

            for (int i = bodies.Count - 1; i >= 0; i--) {
                ((RigidBody)bodies[i]).move2();
            }

            calcVelocities();         

        }

        private void initForces() {
            RigidBody rb;
            for (int i = bodies.Count - 1; i >= 0; i--) {
                rb = bodies[i];
                if (rb.rm != 0 && Form1.GravityEnabled) {
                    rb.force = new Dim2Vector(0F, 8.4F / rb.rm);
                    rb.moment = 0;
                }
                else {
                    rb.force = new Dim2Vector(0F, 0F);
                    rb.moment = 0;
                }
            }
        }

        public void drawAsGroup(int dc) {
            groups.draw(dc);
        }

        public void draw(int dc) {
            for (int i = bodies.Count - 1; i >= 0; i--) {
                ((RigidBody)bodies[i]).draw(dc);
            }
        }

    }
}