
using System;
using System.Collections.Generic;
using System.Drawing;

namespace PDim2Rigid {

    public class RigidBody {

        public RigidBodies owner;
        public int index = -1;
        public Dim2Vector pos = new Dim2Vector();
        public Dim2Vector vel = new Dim2Vector();
        public Dim2Vector force = new Dim2Vector();
        public float theta = 0;
        public float omega = 0;
        public float moment = 0; // [g
        public float rm = 0; // ʂ̋t
        public float rI = 0; // [g̋t
        public float maxR = 0; // ő唼a Փ˔̍
        public Dim2Vector[] verticies; // ϊO
        public Dim2Vector[] tverticies; // ϊ
        public int numVerticies;
        public Color color = Color.Blue;
        public List<HitDataLong> hits = new List<HitDataLong>();

        public RigidGroup group; 

        public Dim2Vector nextVel {
            get {
                return vel + force * (owner.dt * rm);
            }
        }

        public RigidBody(Dim2Vector[] verticies, float idense) {
            this.verticies = verticies;
            numVerticies = verticies.Length;
            tverticies = new Dim2Vector[numVerticies];
            for (int i = numVerticies - 1; i >= 0; i--) {
                tverticies[i] = new Dim2Vector();
            }

            calcData(idense);
        }

        private void calcData(float idense) {

            float m = 0;
            float r1 = 0;
            Dim2Vector center = new Dim2Vector();
            Dim2Vector v1;
            Dim2Vector v2;
            Dim2Vector v3;
            Dim2Vector v4;

            int j;

            for (int i = numVerticies - 1; i >= 0; i--) {

                if (i == 0) {
                    j = numVerticies - 1;
                }
                else
                    j = i - 1;

                v1 = verticies[j];
                v2 = verticies[i];
                v3 = (v1 + v2) * (1.0F / 3.0F);
                v4 = v2 - v1;
                r1 = (v1 % v4) * 0.5F;
                m += r1;
                center += v3 * r1;
            }

            if (m == 0) {
                rm = 0;
                rI = 0;
                return;
            }

            rm = 1 / m;
            center *= rm;
            for (int i = numVerticies - 1; i >= 0; i--) {
                verticies[i] -= center;
            }

            float _I = 0;

            for (int i = numVerticies - 1; i >= 0; i--) {

                r1 = verticies[i].norm();
                maxR = Math.Max(maxR, r1);

                if (i == 0) {
                    j = numVerticies - 1;
                }
                else
                    j = i - 1;

                v1 = verticies[j];
                v2 = verticies[i] - v1;
                r1 = (v1 % v2) * 0.5F
                        * (v1.sqr() / 3F + (v1 * v2) / 2F + v2.sqr() / 3F);
                _I += r1;

            }

            if (_I == 0) {
                rm = 0;
                rI = 0;
                return;
            }

            rI = 1 / _I;

            rm *= idense;
            rI *= idense;

        }

        public Dim2Vector rotateVec(Dim2Vector r1) {
            float s, c;
            c = (float)Math.Cos(theta);
            s = (float)Math.Sin(theta);

            Dim2Vector temp;
            temp.x = r1.x * c - r1.y * s;
            temp.y = r1.x * s + r1.y * c;
            return temp;
        }

        public Dim2Vector transformVec(Dim2Vector r1) {
            float s, c, m13, m23;
            c = (float)Math.Cos(theta);
            s = (float)Math.Sin(theta);
            m13 = pos.x;
            m23 = pos.y;

            Dim2Vector temp;
            temp.x = r1.x * c - r1.y * s + m13;
            temp.y = r1.x * s + r1.y * c + m23;
            return temp;
        }

        public Dim2Vector deTransformVec(Dim2Vector r1) {
            float s, c;
            c = (float)Math.Cos(-theta);
            s = (float)Math.Sin(-theta);

            Dim2Vector temp1, temp2;
            temp1 = r1 - pos;

            temp2.x = temp1.x * c - temp1.y * s;
            temp2.y = temp1.x * s + temp1.y * c;
            return temp2;
        }


        public Boolean IsInPoint(Dim2Vector point) {
            if ((pos - point).sqr() > maxR*maxR)
                return false;
            Dim2Vector l1, l2;
            l1 = point;
            l2 = point;
            l2.x += 5000;
            int j;
            int count = 0;
            for (int i = 0; i < tverticies.Length; i++) {
                j = i + 1;
                if (j == tverticies.Length) j = 0;

                Dim2Vector p;
                if (Dim2Vector.lineHitTest(tverticies[i], tverticies[j],
                    l1, l2, out p)) count++;
            }
            if (count % 2 == 0) {
                return false;
            }
            else {
                return true;
            }
        }

        public Dim2Vector velR(Dim2Vector point) {
            return vel + omega % point;
        }

        public Dim2Vector velRel(RigidBody rb, Dim2Vector r1, Dim2Vector r2) {
            return velR(r1) - rb.velR(r2);
        }

        public Dim2Vector velByImpulse(Dim2Vector argImpulse, Dim2Vector argR) {

            //return velByImpulse(argImpulse, argR, argR);

            float om = rI * (argR.x * argImpulse.y - argR.y * argImpulse.x);
            Dim2Vector v1 = new Dim2Vector(argImpulse.x*rm - argR.y * om,
                argImpulse.y*rm + argR.x * om);
            return v1;
        }

        public Dim2Vector velByImpulse(Dim2Vector argImpulse, Dim2Vector argImpulsePoint,
                Dim2Vector argR) {
            /*
            float mom = argImpulsePoint % argImpulse; // N = r~F

            Dim2Vector v1 = (mom * rI) % argR;
            return argImpulse * rm + v1;*/

            float om = rI * (argImpulsePoint.x * argImpulse.y - argImpulsePoint.y * argImpulse.x);
            Dim2Vector v1 = new Dim2Vector(argImpulse.x*rm - argR.y * om,
                argImpulse.y*rm + argR.x * om);
            return v1;
        }

        public void transForm() {
            float s, c, m13, m23;
            c = (float)Math.Cos(theta);
            s = (float)Math.Sin(theta);
            m13 = pos.x;
            m23 = pos.y;

            for (int i = numVerticies - 1; i >= 0; i--) {
                tverticies[i].x = verticies[i].x * c - verticies[i].y * s + m13;
                tverticies[i].y = verticies[i].x * s + verticies[i].y * c + m23;
            }

        }

        public bool hitTest(RigidBody rb) {

            if ((pos - rb.pos).norm() > maxR + rb.maxR) {
                return false;
            }

            if (rm == 0.0f && rb.rm == 0.0f) {
                return false;
            }

            int k;
            Dim2Vector l11, l12, l21, l22, point;

            int numPoints = 0;
            Dim2Vector[] points = new Dim2Vector[60];
            int[] edges = new int[2 * 60];

            for (int i = numVerticies - 1; i >= 0; i--) {

                k = (i == 0) ? numVerticies - 1 : i - 1;

                l11 = tverticies[k];
                l12 = tverticies[i];

                for (int j = rb.numVerticies - 1; j >= 0; j--) {

                    k = (j == 0) ? rb.numVerticies - 1 : j - 1;

                    l21 = rb.tverticies[k];
                    l22 = rb.tverticies[j];

                    if (Dim2Vector.lineHitTest(l11, l12, l21, l22, out point)) {

                        points[numPoints] = point;
                        edges[2 * numPoints] = (i == 0) ? numVerticies - 1 : i - 1;
                        edges[2 * numPoints + 1] = (j == 0) ? rb.numVerticies - 1
                                : j - 1;
                        numPoints += 1;

                    }
                }
            }

            if (numPoints <= 1) {
                return false;
            }

            float md = -1000000000.0F;
            float a;
            int i1 = 0, i2 = 0;
            l11 = this.pos - rb.pos;

            for (int i = numPoints - 1; i >= 0; i--) {

                for (int j = numPoints - 1; j >= 0; j--) {

                    if (i == j) {
                        continue;
                    }

                    a = l11 * (points[i] - points[j]).cross();
                    if (md <= a) {
                        md = a;
                        i1 = i;
                        i2 = j;
                    }

                }
            }

            if (md == -1000000000.0) {
                return false;
            }

            Dim2Vector n = (points[i1] - points[i2]).cross().nmlz();

            md = 0.0F;
            float md2 = 0.0F;

            md = tverticies[numVerticies - 1] * n;
            for (int i = numVerticies - 2; i >= 0; i--) {
                a = tverticies[i] * n;
                if (md > a) {
                    md = a;
                }
            }

            md2 = rb.tverticies[rb.numVerticies - 1] * n;
            for (int i = rb.numVerticies - 2; i >= 0; i--) {
                a = rb.tverticies[i] * n;
                if (md2 < a) {
                    md2 = a;
                }
            }

            if (md2 - md < 0) {
                Console.Write("aaaaaaaaaaaa");

            }


            md = Math.Max(md2 - md - 0.3F, 0);

            if (n.norm() == 0.0f) {
                n = (pos - rb.pos).nmlz();
                if (n.norm() == 0.0f) {
                    n = new Dim2Vector(1.0f, 0.0f);
                }
            }

            HitDataShort hd2 = new HitDataShort();
            hd2.n = n;
            hd2.fn = n.cross();
            hd2.h = md;
            hd2.impulseConst = owner.ImpulseConst;
            hd2.fricConst = owner.FricConst;
            hd2.hdl = new HitDataLong();
            hd2.hdl.rb1 = this;
            hd2.hdl.rb2 = rb;
            hd2.hdl.edge1 = edges[2 * i1 + 0];
            hd2.hdl.edge2 = edges[2 * i1 + 1];
            hd2.r1 = points[i1] - this.pos;
            hd2.r2 = points[i1] - rb.pos;
            owner.addHitData(hd2);

            hd2 = new HitDataShort();
            hd2.n = n;
            hd2.fn = n.cross();
            hd2.h = md;
            hd2.impulseConst = owner.ImpulseConst;
            hd2.fricConst = owner.FricConst;
            hd2.hdl = new HitDataLong();
            hd2.hdl.rb1 = this;
            hd2.hdl.rb2 = rb;
            hd2.hdl.edge1 = edges[2 * i2 + 0];
            hd2.hdl.edge2 = edges[2 * i2 + 1];
            hd2.r1 = points[i2] - this.pos;
            hd2.r2 = points[i2] - rb.pos;
            owner.addHitData(hd2);

            return true;
        }

        public void move1() {

            for (int i = hits.Count - 1; i >= 0; i--) {
                if (hits[i].decAge()) {
                    hits.RemoveAt(i);
                }
            }

            pos += vel * owner.dt;
            theta += omega * owner.dt;

            transForm();

        }

        public void move2() {

            vel += force * (owner.dt * rm);
            omega += moment * owner.dt * rI;

        }

        public void draw(int dc) {

            int i;

            Point[] points;
            points = new Point[numVerticies];
            for (i = numVerticies - 1; i >= 0; i--) {
                points[i].X = (int)tverticies[i].x;
                points[i].Y = (int)tverticies[i].y;
            }

            GDIFunctions.SelectObject(dc, GDIFunctions.GetStockObject(GDIFunctions.DC_PEN));
            GDIFunctions.SetDCPenColor(dc, GDIFunctions.ColorToInt(Color.Black));
            GDIFunctions.SelectObject(dc, GDIFunctions.GetStockObject(GDIFunctions.DC_BRUSH));
            GDIFunctions.SetDCBrushColor(dc, GDIFunctions.ColorToInt(color));

            GDIFunctions.Polygon(dc, points, points.Length);
        }

    }



}
